# __future__ import needed for classmethod factory functions; should be dropped
# with py 3.10.
from __future__ import annotations

import re
from dataclasses import asdict, dataclass, field
from pathlib import Path, PurePosixPath
from urllib.parse import quote_plus

import toml
from docker.models.containers import Container

from .core import LOGGER
from .util import docker_mount_for, get_docker_env, is_volume_readonly

MAIN_PROFILE_NAME = "default"

CONTAINER_PREFIX = "aiidalab_"

DEFAULT_PORT = 8888

# Regular expression for valid container and image names as of Docker version 20.10.11:
# [a-zA-Z0-9][a-zA-Z0-9_.-]+
_REGEX_VALID_IMAGE_NAMES = r"[a-zA-Z0-9][a-zA-Z0-9_.-]+"
# For profiles in addition we do not allow underscores to avoid potential issues with moving
# to a docker-compose based implementation of aiidalab-launch in the future.
_REGEX_VALID_PROFILE_NAMES = r"[a-zA-Z0-9][a-zA-Z0-9.-]+"


def _default_port() -> int:  # explicit function required to enable test patching
    return DEFAULT_PORT


_DEFAULT_IMAGE = "aiidalab/full-stack:latest"


def _valid_volume_name(source: str) -> None:
    # We do not allow relative paths so if the path is not absolute,
    # we assume volume mount, whose name is restricted by Docker.
    if not Path(source).is_absolute() and not re.fullmatch(
        _REGEX_VALID_IMAGE_NAMES, source
    ):
        raise ValueError(
            f"Invalid extra mount volume name '{source}'. Use absolute path for bind mounts."
        )


def _get_configured_host_port(container: Container) -> int | None:
    try:
        host_config = container.attrs["HostConfig"]
        return int(host_config["PortBindings"]["8888/tcp"][0]["HostPort"]) or None
    except (KeyError, IndexError, ValueError):
        pass
    return None


def _get_aiidalab_default_apps(container: Container) -> list:
    try:
        return get_docker_env(container, "AIIDALAB_DEFAULT_APPS").split()
    except KeyError:
        return []


@dataclass
class Profile:
    name: str = MAIN_PROFILE_NAME
    port: int | None = field(default_factory=_default_port)
    default_apps: list[str] = field(default_factory=lambda: ["aiidalab-widgets-base"])
    system_user: str = "jovyan"
    image: str = _DEFAULT_IMAGE
    home_mount: str | None = None
    extra_mounts: set[str] = field(default_factory=set)

    def __post_init__(self):
        if (
            not re.fullmatch(_REGEX_VALID_PROFILE_NAMES, self.name)
            or quote_plus(self.name) != self.name
        ):
            raise ValueError(
                f"Invalid profile name '{self.name}'. The profile name must be "
                "composed of the following characters [a-zA-Z0-9.-] and must "
                "start with an alphanumeric character."
            )
        if self.home_mount is None:
            self.home_mount = f"{CONTAINER_PREFIX}{self.name}_home"

        _valid_volume_name(self.home_mount)

        # Normalize extra mount mode to be "rw" by default
        # so that we match Docker default but are explicit.
        for extra_mount in self.extra_mounts:
            self.parse_extra_mount(extra_mount)
            if len(extra_mount.split(":")) == 2:
                self.extra_mounts.remove(extra_mount)
                self.extra_mounts.add(f"{extra_mount}:rw")

        if (
            self.image.split(":")[0] == "aiidalab/full-stack"
            and self.system_user != "jovyan"
        ):
            LOGGER.warning(
                "Resetting the system user may create issues for this image!"
            )

    def container_name(self) -> str:
        return f"{CONTAINER_PREFIX}{self.name}"

    def parse_extra_mount(
        self, extra_mount: str
    ) -> tuple[Path, PurePosixPath, str | None]:
        fields = extra_mount.split(":")
        if len(fields) < 2 or len(fields) > 3:
            raise ValueError(f"Invalid extra mount option '{extra_mount}'")

        source, target = fields[:2]
        _valid_volume_name(source)

        source_path, target_path = Path(source), PurePosixPath(target)
        # Unlike for home_mount, we will not auto-create missing
        # directories for extra mounts.
        if source_path.is_absolute() and not source_path.exists():
            raise ValueError(f"Directory '{source}' does not exist")

        # By default, extra mounts are writeable
        mode = fields[2] if len(fields) == 3 else "rw"
        if mode not in ("ro", "rw"):
            raise ValueError(f"Invalid extra mount mode '{mode}'")

        return source_path, target_path, mode

    def conda_volume_name(self) -> str:
        return f"{self.container_name()}_conda"

    def environment(self, jupyter_token: str) -> dict:
        return {
            "AIIDALAB_DEFAULT_APPS": " ".join(self.default_apps),
            "JUPYTER_TOKEN": str(jupyter_token),
            "SYSTEM_USER": self.system_user,
            "NB_USER": self.system_user,
        }

    def dumps(self) -> str:
        return toml.dumps({k: v for k, v in asdict(self).items() if k != "name"})

    @classmethod
    def loads(cls, name: str, s: str) -> Profile:
        params = toml.loads(s)
        params["extra_mounts"] = set(params["extra_mounts"])
        return cls(name=name, **params)

    @classmethod
    def from_container(cls, container: Container) -> Profile:
        profile_name = re.sub(re.escape(CONTAINER_PREFIX), "", container.name)
        if not profile_name:
            raise RuntimeError(
                f"Container {container.id} does not appear to be an AiiDAlab container."
            )

        system_user = get_docker_env(container, "SYSTEM_USER")

        image_tag = (
            _DEFAULT_IMAGE
            if _DEFAULT_IMAGE in container.image.tags
            else container.image.tags[0]
        )

        extra_destinations: list[PurePosixPath] = [
            PurePosixPath(mount["Destination"])
            for mount in container.attrs["Mounts"]
            if mount["Destination"]
            not in (f"/home/{system_user}/.conda", f"/home/{system_user}")
        ]
        extra_mounts: set[str] = set()
        for dst in extra_destinations:
            src = docker_mount_for(container, dst)
            if is_volume_readonly(container, dst):
                extra_mounts.add(":".join([str(src), str(dst), "ro"]))
            else:
                extra_mounts.add(":".join([str(src), str(dst), "rw"]))

        return Profile(
            name=profile_name,
            port=_get_configured_host_port(container),
            default_apps=_get_aiidalab_default_apps(container),
            home_mount=str(
                docker_mount_for(container, PurePosixPath("/", "home", system_user))
            ),
            image=image_tag,
            system_user=system_user,
            extra_mounts=extra_mounts,
        )



# __future__ import needed for classmethod factory functions; should be dropped
# with py 3.10.
from __future__ import annotations

from copy import deepcopy
from dataclasses import asdict, dataclass, field
from pathlib import Path
from uuid import uuid4

import toml

from .profile import Profile

MAIN_PROFILE_NAME = "default"


@dataclass
class Config:
    profiles: list[Profile] = field(default_factory=lambda: [Profile()])
    default_profile: str = MAIN_PROFILE_NAME

    # The configuration is always stored to disk beginning with version
    # 2022.1012, which means we assume that if no configuration is stored
    # we cannot make any assumptions about the latest applicable version.
    version: str | None = None

    @classmethod
    def loads(cls, blob: str) -> Config:
        loaded_config = toml.loads(blob)
        config = deepcopy(loaded_config)
        config["profiles"] = []
        for name, profile in loaded_config.pop("profiles", dict()).items():
            extra_mounts = (
                set(profile.pop("extra_mounts")) if "extra_mounts" in profile else set()
            )
            config["profiles"].append(
                Profile(name=name, extra_mounts=extra_mounts, **profile)
            )
        return cls(**config)

    def dumps(self) -> str:
        config = asdict(self)
        config["profiles"] = {
            profile.pop("name"): profile for profile in config.pop("profiles", [])
        }
        return toml.dumps(config)

    @classmethod
    def load(cls, path: Path) -> Config:
        return cls.loads(path.read_text())

    def save(self, path: Path, safe: bool = True) -> None:
        path.parent.mkdir(exist_ok=True, parents=True)
        if safe:
            path_tmp = path.with_suffix(f".{uuid4()!s}")
            path_tmp.write_text(self.dumps())
            path_tmp.replace(path)
        else:
            path.write_text(self.dumps())

    def get_profile(self, name: str) -> Profile:
        for profile in self.profiles:
            if profile.name == name:
                return profile
        raise ValueError(f"Did not find profile with name '{name}'.")



# __future__ import needed for classmethod factory functions; should be dropped
# with py 3.10.
import logging

APPLICATION_ID = "org.aiidalab.aiidalab_launch"

LOGGER = logging.getLogger(APPLICATION_ID.split(".")[-1])



#!/usr/bin/env python
"""Tool to launch and manage AiiDAlab instances with docker.

Authors:
    * Carl Simon Adorf <simon.adorf@epfl.ch>
"""
from dataclasses import dataclass, field
from pathlib import Path

import click
import docker
from packaging.version import parse

from .config import Config
from .core import APPLICATION_ID
from .instance import AiidaLabInstance
from .profile import Profile
from .util import get_docker_client
from .version import __version__


def _application_config_path():
    return Path(click.get_app_dir(APPLICATION_ID)) / "config.toml"


def _load_config():
    try:
        return Config.load(_application_config_path())
    except FileNotFoundError:
        return Config()


@dataclass
class ApplicationState:

    config_path: Path = field(default_factory=_application_config_path)
    config: Config = field(default_factory=_load_config)
    docker_client: docker.DockerClient = field(default_factory=get_docker_client)

    def save_config(self):
        self.config.save(self.config_path)

    def _apply_migration_null(self):
        # Since there is no config file on disk, we can assume that if at all,
        # there is only the default profile present.
        assert len(self.config.profiles) == 1
        assert self.config.profiles[0].name == "default"

        default_profile = self.config.profiles[0]
        instance = AiidaLabInstance(client=self.docker_client, profile=default_profile)

        # Default home bind mount path up until version 2022.1011.
        home_bind_mount_path = Path.home() / "aiidalab"

        if instance.container:
            # There is already a container present, use previously used profile.
            self.config.profiles[0] = Profile.from_container(instance.container)

        elif home_bind_mount_path.exists():
            # Using ~/aiidalab as home directory mount point, since the
            # directory exists. The default mount point was changed to be a
            # docker volume after version 2022.1011 to address issue
            # https://github.com/aiidalab/aiidalab-launch/issues/72.
            self.config.profiles[0].home_mount = str(home_bind_mount_path)

    def apply_migrations(self):
        config_changed = False

        # No config file saved to disk.
        if not self.config_path.is_file():
            self._apply_migration_null()
            config_changed = True

        # No version string stored in config.
        if self.config.version != str(parse(__version__)):
            self.config.version = str(parse(__version__))
            config_changed = True

        # Write any changes back to disk.
        if config_changed:
            self.save_config()



# __future__ import needed for classmethod factory functions; should be dropped
# with py 3.10.
from __future__ import annotations

import asyncio
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path, PurePosixPath
from secrets import token_hex
from shutil import rmtree
from time import time
from typing import Any, AsyncGenerator, Generator

import docker
from docker.models.containers import Container

from .core import LOGGER
from .profile import Profile
from .util import _async_wrap_iter, docker_mount_for, get_docker_env


def _get_host_ports(container: Container) -> Generator[int, None, None]:
    try:
        ports = container.attrs["NetworkSettings"]["Ports"]
        yield from (int(i["HostPort"]) for i in ports["8888/tcp"])
    except KeyError:
        pass


class FailedToWaitForServices(RuntimeError):
    pass


class RequiresContainerInstance(RuntimeError):
    """Raised when trying to perform operation that requires a container instance."""


class NoHostPortAssigned(RuntimeError):
    """Raised when then trying to obtain the instance URL, but there is not host port."""


@contextmanager
def _async_logs(
    container: Container,
) -> Generator[AsyncGenerator[Any, None], None, None]:
    logs = container.logs(stream=True, follow=False)
    try:
        yield _async_wrap_iter(logs)
    finally:
        logs.close()


@dataclass
class AiidaLabInstance:
    class AiidaLabInstanceStatus(Enum):
        UNKNOWN = auto()
        CREATED = auto()
        DOWN = auto()
        UP = auto()
        EXITED = auto()
        STARTING = auto()

    client: docker.DockerClient
    profile: Profile
    _image: docker.models.images.Image = None
    _container: Container = None
    _protocol: str = "http"

    def _get_image(self) -> docker.models.images.Image | None:
        try:
            return self.client.images.get(self.profile.image)
        except docker.errors.ImageNotFound:
            return None

    def _get_container(self) -> Container | None:
        try:
            return self.client.containers.get(self.profile.container_name())
        except docker.errors.NotFound:
            return None

    def __post_init__(self) -> None:
        self._image = self._get_image()
        self._container = self._get_container()

    @property
    def image(self) -> docker.models.images.Image | None:
        return self._image

    @property
    def container(self) -> Container | None:
        if self._container is None:
            self._container = self._get_container()
        return self._container

    def _requires_container(self) -> Container:
        container = self.container
        if container is None:
            raise RequiresContainerInstance
        return container

    def _conda_mount(self) -> docker.types.Mount:
        return docker.types.Mount(
            target=f"/home/{self.profile.system_user}/.conda",
            source=self.profile.conda_volume_name(),
        )

    def _home_mount(self) -> docker.types.Mount:
        assert self.profile.home_mount is not None
        home_mount_path = Path(self.profile.home_mount)
        return docker.types.Mount(
            target=f"/home/{self.profile.system_user}",
            source=self.profile.home_mount,
            type="bind" if home_mount_path.is_absolute() else "volume",
        )

    def _extra_mount(self, extra_mount: str) -> docker.types.Mount:
        source_path, target_path, mode = self.profile.parse_extra_mount(extra_mount)
        return docker.types.Mount(
            target=str(target_path),
            source=str(source_path),
            read_only=True if mode == "ro" else False,
            type="bind" if source_path.is_absolute() else "volume",
        )

    def _mounts(self) -> Generator[docker.types.Mount, None, None]:
        yield self._conda_mount()
        if self.profile.home_mount:
            yield self._home_mount()
        for extra_mount in self.profile.extra_mounts:
            yield self._extra_mount(extra_mount)

    def configuration_changes(self) -> Generator[str, None, None]:
        self._requires_container()
        assert self.container is not None
        assert self.image is not None

        if self.container.image.id != self.image.id:
            yield "Image has changed."

        try:
            for mount in self._mounts():
                if (
                    str(docker_mount_for(self.container, mount["Target"]))
                    != mount["Source"]
                ):
                    raise ValueError
        except ValueError:
            yield "Mount configuration has changed."

        if self.profile != Profile.from_container(self.container):
            yield "Profile configuration has changed."

    def pull(self) -> docker.models.images.Image | None:
        try:
            image = self.client.images.pull(self.profile.image)
            LOGGER.info(f"Pulled image: {image}")
            self._image = image
            return image
        except docker.errors.NotFound:
            LOGGER.warn(f"Unable to pull image: {self.profile.image}")
            return None

    def _ensure_home_mount_exists(self) -> None:
        if self.profile.home_mount:
            home_mount_path = Path(self.profile.home_mount)
            if home_mount_path.is_absolute():
                LOGGER.info(
                    f"Ensure home mount point ({self.profile.home_mount}) exists."
                )
                home_mount_path.mkdir(exist_ok=True, parents=True)

    def create(self) -> Container:
        assert self._container is None
        self._ensure_home_mount_exists()
        self._container = self.client.containers.create(
            image=(self.image or self.pull()),
            name=self.profile.container_name(),
            environment=self.profile.environment(jupyter_token=token_hex(32)),
            mounts=list(self._mounts()),
            ports={"8888/tcp": self.profile.port},
        )
        return self._container

    def recreate(self) -> None:
        self._requires_container()
        assert self.container is not None
        self.remove()
        self.create()

    def start(self) -> None:
        self._ensure_home_mount_exists()
        LOGGER.info(f"Starting container '{self.profile.container_name()}'...")
        (self.container or self.create()).start()
        assert self.container is not None
        LOGGER.info(f"Started container: {self.container.name} ({self.container.id}).")
        self._run_post_start()

    def restart(self) -> None:
        self._requires_container()
        assert self.container is not None
        self.container.restart()
        self._run_post_start()

    def _run_post_start(self) -> None:
        assert self.container is not None
        LOGGER.debug("Run post-start commands.")

        LOGGER.debug("Ensure ~/.conda directory is owned by the system user.")
        exit_code, _ = self.container.exec_run(
            f"chown -R 1000:1000 /home/{self.profile.system_user}/.conda",
            privileged=True,
        )
        if exit_code != 0:
            LOGGER.warning(
                "Failed to ensure ~/.conda directory is owned by the system user."
            )
        else:
            LOGGER.debug("The ~/.conda directory is owned by the system user.")

    def stop(self, timeout: float | None = None) -> None:
        self._requires_container()
        assert self.container is not None
        try:
            self.container.stop(timeout=timeout)
        except AttributeError:
            raise RuntimeError("no container")

    def remove(self, conda: bool = False, data: bool = False) -> None:
        # Remove container
        if self.container:
            self.container.remove()
            self._container = None

        # Remove conda volume
        if conda:
            try:
                self.client.volumes.get(self.profile.conda_volume_name()).remove()
            except docker.errors.NotFound:  # already removed
                LOGGER.debug(
                    f"Failed to remove conda volume '{self.profile.conda_volume_name()}', likely already removed."
                )
            except Exception as error:  # unexpected error
                raise RuntimeError(f"Failed to remove conda volume: {error}")

        if data and self.profile.home_mount:
            # Remove home volume
            home_mount_path = PurePosixPath(self.profile.home_mount)
            try:
                if home_mount_path.is_absolute():
                    rmtree(home_mount_path)
                else:
                    self.client.volumes.get(str(home_mount_path)).remove()
            except docker.errors.NotFound:
                pass  # already removed
            except Exception as error:  # unexpected error
                raise RuntimeError(f"Failed to remove home volume: {error}")

    def logs(
        self, stream: bool = False, follow: bool = False
    ) -> docker.types.daemon.CancellableStream | str:
        if self.container is None:
            raise RuntimeError("Instance was not created.")
        return self.container.logs(stream=stream, follow=follow)

    def exec_create(self, cmd: str, privileged: bool = False) -> str:
        LOGGER.info(f"Executing: {' '.join(cmd)}")
        if self.container is None:
            raise RuntimeError("Instance was not created.")

        try:
            return self.client.api.exec_create(
                self.container.id,
                cmd,
                user="root" if privileged else self.profile.system_user,
                workdir=None if privileged else f"/home/{self.profile.system_user}",
            )["Id"]
        except docker.errors.APIError:
            raise RuntimeError(
                f"Unable to send command to container '{self.container.id}'."
            )

    async def echo_logs(self) -> None:
        assert self.container is not None
        with _async_logs(self.container) as logs:
            async for chunk in logs:
                LOGGER.debug(f"{self.container.id}: {chunk.decode('utf-8').strip()}")

    @staticmethod
    async def _init_scripts_finished(container: Container) -> None:
        loop = asyncio.get_event_loop()
        LOGGER.info("Waiting for init services to finish...")
        result = await loop.run_in_executor(
            None, container.exec_run, "wait-for-services"
        )
        if result.exit_code in (126, 127):
            LOGGER.debug("Container does not support wait-for-services script.")
        elif result.exit_code != 0:
            raise FailedToWaitForServices(
                "Failed to check for init processes to complete."
            )
        LOGGER.info("Init services finished.")

    @staticmethod
    async def _notebook_service_online(container: Container) -> str:
        loop = asyncio.get_event_loop()
        LOGGER.info("Waiting for notebook service to become reachable...")
        assumed_protocol = "http"
        while True:
            try:
                LOGGER.debug("Curl notebook...")
                result = await loop.run_in_executor(
                    None,
                    container.exec_run,
                    f"curl --fail-early --fail --silent --max-time 1.0 {assumed_protocol}://localhost:8888",
                )
                if result.exit_code == 0:
                    LOGGER.info("Notebook service reachable.")
                    return assumed_protocol
                elif result.exit_code in (7, 28):
                    await asyncio.sleep(2)  # jupyter not yet reachable
                    continue
                elif result.exit_code == 56 and assumed_protocol == "http":
                    assumed_protocol = "https"
                    LOGGER.info("Trying to connect via HTTPS.")
                    continue
                elif result.exit_code == 60:
                    LOGGER.info("Notebook service reachable.")
                    LOGGER.warning("Could not authenticate HTTPS certificate.")
                    return assumed_protocol
                else:
                    raise FailedToWaitForServices("Failed to reach notebook service.")
            except docker.errors.APIError:
                LOGGER.error("Failed to reach notebook service. Aborting.")
                raise FailedToWaitForServices(
                    "Failed to reach notebook service (unable to reach container."
                )

    @staticmethod
    async def _host_port_assigned(container: Container) -> None:
        LOGGER.debug("Waiting for host port to be assigned...")
        while True:
            container.reload()
            if any(_get_host_ports(container)):
                LOGGER.debug("Host port assigned.")
                break
            await asyncio.sleep(1)

    async def wait_for_services(self) -> None:
        container = self._requires_container()
        LOGGER.info(f"Waiting for services to come up ({container.id})...")
        start = time()
        _, self._protocol, _ = await asyncio.gather(
            self._init_scripts_finished(container),
            self._notebook_service_online(container),
            self._host_port_assigned(container),
        )
        LOGGER.info(
            f"Services came up after {time() - start:.1f} seconds ({container.id})."
        )

    async def status(self, timeout: float | None = 5.0) -> AiidaLabInstanceStatus:
        if self.container:
            self.container.reload()
            await asyncio.sleep(0)
            if self.container.status == "running":
                try:
                    await asyncio.wait_for(self.wait_for_services(), timeout=timeout)
                except asyncio.TimeoutError:
                    return self.AiidaLabInstanceStatus.STARTING
                except RuntimeError:
                    return self.AiidaLabInstanceStatus.UNKNOWN
                else:
                    return self.AiidaLabInstanceStatus.UP
            elif self.container.status == "created":
                return self.AiidaLabInstanceStatus.CREATED
            elif self.container and self.container.status == "exited":
                return self.AiidaLabInstanceStatus.EXITED
        return self.AiidaLabInstanceStatus.DOWN

    def host_ports(self) -> list[int]:
        self._requires_container()
        assert self.container is not None
        self.container.reload()
        return list(_get_host_ports(self.container))

    def url(self) -> str:
        self._requires_container()
        assert self.container is not None
        self.container.reload()
        host_ports = list(_get_host_ports(self.container))
        if len(host_ports) > 0:
            jupyter_token = get_docker_env(self.container, "JUPYTER_TOKEN")
            return (
                f"{self._protocol}://localhost:{host_ports[0]}/?token={jupyter_token}"
            )
        else:
            raise NoHostPortAssigned(self.container.id)



#!/usr/bin/env python
"""Tool to launch and manage AiiDAlab instances with docker.

Authors:
    * Carl Simon Adorf <simon.adorf@epfl.ch>
"""
import asyncio
import getpass
import logging
import socket
import sys
from pathlib import Path
from textwrap import wrap

import click
import docker
from packaging.version import parse
from tabulate import tabulate

from .application_state import ApplicationState
from .core import LOGGER
from .instance import AiidaLabInstance
from .profile import DEFAULT_PORT, Profile
from .util import confirm_with_value, get_latest_version, spinner, webbrowser_available
from .version import __version__

MSG_MOUNT_POINT_CONFLICT = """Warning: There is at least one other running
instance that has the same home mount point ('{home_mount}') as the instance
you are currently trying to start. Are you sure you want to continue? This may
lead to data corruption."""


MSG_STARTUP = """Open the following URL to access AiiDAlab:

  {url}

Home mounted: {home_mount} -> /home/{system_user}"""


MSG_STARTUP_SSH = """
Unable to detect a web browser which indicates that you might be running
AiiDAlab on a remote machine. If this is the case, consider to create an SSH
tunnel to access AiiDAlab on your local computer. For this, run a command
similar to

  ssh {user}@{hostname} -NfL {port}:localhost:{port}

on your local computer, then open AiiDAlab on your local computer at

  {url}

See https://github.com/aiidalab/aiidalab-launch/blob/main/ssh-forward.md for
more detailed instructions on SSH port forwarding.

Home mounted: {home_mount} -> /home/{system_user}"""

MSG_EXTRA_VOLUME = "Extra volume mounted: {source} -> {target} {mode}"


LOGGING_LEVELS = {
    0: logging.ERROR,
    1: logging.WARN,
    2: logging.INFO,
    3: logging.DEBUG,
}  #: a mapping of `verbose` option counts to logging levels


pass_app_state = click.make_pass_decorator(ApplicationState, ensure=True)


def exception_handler(exception_type, exception, traceback):  # noqa: U100
    click.echo(f"Unexpected {exception_type.__name__}: {exception}", err=True)
    click.echo(
        "Use verbose mode `aiidalab-launch --verbose` to see full stack trace", err=True
    )


def with_profile(cmd):
    def callback(ctx, param, value):  # noqa: U100
        app_state = ctx.ensure_object(ApplicationState)
        name = value or app_state.config.default_profile
        LOGGER.info(f"Using profile: {name}")
        return app_state.config.get_profile(name)

    return click.option(
        "-p", "--profile", help="Select profile to use.", callback=callback
    )(cmd)


@click.group()
@click.option(
    "-v",
    "--verbose",
    count=True,
    help=(
        "Increase the output verbosity of the launcher. "
        "Use '-vv' or '-vvv' for even more verbose output"
    ),
)
@pass_app_state
def cli(app_state, verbose):
    # Use the verbosity count to determine the logging level...
    logging.basicConfig(
        level=LOGGING_LEVELS[verbose] if verbose in LOGGING_LEVELS else logging.DEBUG
    )
    if verbose > 0:
        click.secho(
            f"Verbose logging is enabled. "
            f"(LEVEL={logging.getLogger().getEffectiveLevel()})",
            fg="yellow",
            err=True,
        )

    # Hide stack trace by default.
    if verbose == 0:
        sys.excepthook = exception_handler

    LOGGER.info(f"Configuration file path: {app_state.config_path}")

    latest_version = get_latest_version(timeout=0.1)
    if latest_version and latest_version > parse(__version__):
        click.secho(
            f"A new version of aiidalab-launch is available: {latest_version} (installed: {__version__})",
            fg="yellow",
        )
        if "pipx" in __file__:
            click.secho("Run `pipx upgrade aiidalab-launch` to update.", fg="yellow")

    # Apply migrations
    app_state.apply_migrations()


@cli.command()
def version():
    """Show the version of aiidalab-launch."""
    click.echo(click.style(f"AiiDAlab Launch {__version__}", bold=True))


@cli.group()
def profiles():
    """Manage AiiDAlab profiles."""
    pass


@profiles.command("list")
@pass_app_state
def list_profiles(app_state):
    """List all configured AiiDAlab profiles.

    The default profile is shown in bold.
    """
    default_profile = app_state.config.default_profile
    click.echo(
        "\n".join(
            [
                click.style(
                    profile.name + (" *" if profile.name == default_profile else ""),
                    bold=profile.name == default_profile,
                )
                for profile in app_state.config.profiles
            ]
        )
    )


@profiles.command("show")
@click.argument("profile")
@pass_app_state
def show_profile(app_state, profile):
    """Show an AiiDAlab profile configuration."""
    click.echo(app_state.config.get_profile(profile).dumps(), nl=False)


@profiles.command("add")
@click.argument("profile")
@click.option(
    "--port",
    type=click.IntRange(min=1, max=65535),
    help=(
        "Specify port on which this instance will be exposed. The default port "
        "is chosen such that it does not conflict with any currently configured "
        "profiles."
    ),
)
@click.option(
    "--home-mount",
    type=click.Path(file_okay=False),
    help="Specify the path to be mounted as home directory.",
)
@pass_app_state
@click.pass_context
def add_profile(ctx, app_state, port, home_mount, profile):
    """Add a new AiiDAlab profile to the configuration."""
    try:
        app_state.config.get_profile(profile)
    except ValueError:
        pass
    else:
        raise click.ClickException(f"Profile with name '{profile}' already exists.")

    # Determine next available port or use the one provided by the user.
    configured_ports = [prof.port for prof in app_state.config.profiles if prof.port]
    port = port or (max(configured_ports, default=-1) + 1) or DEFAULT_PORT

    try:
        new_profile = Profile(
            name=profile,
            port=port,
            home_mount=home_mount,
        )
    except ValueError as error:  # invalid profile name
        raise click.ClickException(error)

    app_state.config.profiles.append(new_profile)
    app_state.save_config()
    click.echo(f"Added profile '{profile}'.")
    if click.confirm("Do you want to edit it now?", default=True):
        ctx.invoke(edit_profile, profile=profile)


@profiles.command("remove")
@click.argument("profile")
@click.option("--yes", is_flag=True, help="Do not ask for confirmation.")
@click.option("-f", "--force", is_flag=True, help="Proceed, ignoring any warnings.")
@pass_app_state
def remove_profile(app_state, profile, yes, force):
    """Remove an AiiDAlab profile from the configuration."""
    try:
        profile = app_state.config.get_profile(profile)
    except ValueError:
        raise click.ClickException(f"Profile with name '{profile}' does not exist.")
    else:
        if not force:
            instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
            status = asyncio.run(instance.status())
            if status not in (
                instance.AiidaLabInstanceStatus.DOWN,
                instance.AiidaLabInstanceStatus.CREATED,
                instance.AiidaLabInstanceStatus.EXITED,
            ):
                raise click.ClickException(
                    f"The instance associated with profile '{profile.name}' "
                    "is still running. Use the -f/--force option to remove the "
                    "profile anyways."
                )

        if yes or click.confirm(
            f"Are you sure you want to remove profile '{profile.name}'?"
        ):
            app_state.config.profiles.remove(profile)
            app_state.save_config()
            click.echo(f"Removed profile with name '{profile.name}'.")


@profiles.command("edit")
@click.argument("profile")
@pass_app_state
def edit_profile(app_state, profile):
    """Edit an AiiDAlab profile configuration."""
    current_profile = app_state.config.get_profile(profile)
    profile_edit = click.edit(current_profile.dumps(), extension=".toml")
    if profile_edit:
        new_profile = Profile.loads(profile, profile_edit)
        if new_profile != current_profile:
            app_state.config.profiles.remove(current_profile)
            app_state.config.profiles.append(new_profile)
            app_state.save_config()
            return
    click.echo("No changes.")


@profiles.command("set-default")
@click.argument("profile")
@pass_app_state
def set_default_profile(app_state, profile):
    """Set an AiiDAlab profile as default."""
    try:
        app_state.config.get_profile(profile)
    except ValueError:
        raise click.ClickException(f"A profile with name '{profile}' does not exist.")
    else:
        app_state.config.default_profile = profile
        app_state.save_config()
        click.echo(f"Set default profile to '{profile}'.")


async def _find_mount_point_conflict(client, profile, other_profiles):
    """Find running instances with the same home mount point.

    To protect users from inadvertently starting a second profile with the same
    home mount point. Running two containers with the same home mount point has
    potential for data corruption.
    """
    for other_profile in other_profiles:
        if (
            other_profile != profile
            and Path(other_profile.home_mount).resolve()
            == Path(profile.home_mount).resolve()
        ):
            status = await AiidaLabInstance(
                client=client, profile=other_profile
            ).status()
            if status is not AiidaLabInstance.AiidaLabInstanceStatus.DOWN:
                yield other_profile


async def _async_start(
    app_state, profile, restart, wait, pull, no_browser, show_ssh_help, force
):
    # Check for potential mount point conflicts.
    instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)

    if not force:
        conflict = False
        with spinner("Check for potential conflicts...", delay=0.1):
            async for p in _find_mount_point_conflict(
                app_state.docker_client,
                profile,
                app_state.config.profiles,
            ):
                conflict = True
                break

        if conflict:
            msg_warn = MSG_MOUNT_POINT_CONFLICT.format(home_mount=profile.home_mount)
            click.confirm(
                click.style("\n".join(wrap(msg_warn)), fg="yellow"),
                abort=True,
            )

    # Obtain image (either via pull or local).
    if pull:
        msg = (
            f"Downloading image '{instance.profile.image}', this may take a while..."
            if instance.image is None
            else f"Downloading latest version of '{instance.profile.image}'..."
        )
        with spinner(msg):
            instance.pull()

    if instance.image is None:
        raise click.ClickException(
            f"Unable to find image '{profile.image}'. "
            "Try to use '--pull' to pull the image prior to start."
        )

    # Check if the container configuration has changed.
    if instance.container:
        configuration_changed = any(instance.configuration_changes())
    else:
        configuration_changed = False

    try:

        InstanceStatus = instance.AiidaLabInstanceStatus  # local alias for brevity

        status = await instance.status()

        # Container needs to be started.
        if status in (
            InstanceStatus.DOWN,
            InstanceStatus.CREATED,
            InstanceStatus.EXITED,
        ):
            if configuration_changed:
                with spinner("Recreating container..."):
                    instance.recreate()
            with spinner("Starting container..."):
                instance.start()

        # Container is already up.
        elif status is InstanceStatus.UP and restart:
            with spinner("Restarting container..."):
                if configuration_changed:
                    instance.stop()
                    instance.recreate()
                    instance.start()
                else:
                    instance.restart()
        elif status is InstanceStatus.UP and not restart:
            if configuration_changed:
                click.secho(
                    "Container is already running, however the configuration "
                    "has changed. A restart with --restart is recommended.",
                    fg="yellow",
                )
            else:
                click.echo(
                    "Container was already running, use --restart to restart it.",
                    err=True,
                )

        # Container is already starting.
        elif status is InstanceStatus.STARTING:
            click.echo("Container is already starting up...", err=True)

        # Unknown condition.
        else:
            raise RuntimeError(
                "Container already exists, but failed to determine status."
            )

    except docker.errors.APIError as error:
        LOGGER.debug(f"Error during startup: {error}")
        if instance.profile.port and "port is already allocated" in str(error):
            raise click.ClickException(
                f"Port {instance.profile.port} is already allocated, choose another port "
                f"for example, by editing the profile: aiidalab-launch profiles edit {instance.profile.name}"
            )
        raise click.ClickException("Startup failed due to an unexpected error.")
    except asyncio.TimeoutError:
        raise click.ClickException(
            "AiiDAlab instance did not start up within the excepted wait period."
        )
    except Exception as error:
        raise click.ClickException(f"Unknown error occurred: {error}")
    else:
        if wait:
            try:
                with spinner("Waiting for AiiDAlab instance to get ready..."):
                    echo_logs = asyncio.create_task(instance.echo_logs())
                    await asyncio.wait_for(instance.wait_for_services(), timeout=wait)
                    echo_logs.cancel()
                    LOGGER.debug("AiiDAlab instance ready.")
            except asyncio.TimeoutError:
                raise click.ClickException(
                    f"AiiDAlab instance did not start up within the provided wait period ({wait})."
                )
            except RuntimeError:
                raise click.ClickException(
                    "The AiiDAlab instance failed to start. You can inspect "
                    "the container output logs with "
                    f"'aiidalab-launch logs -p {instance.profile.name}' "
                    "and increase the output verbosity with "
                    "'aiidalab-launch -vvv start'."
                )
            LOGGER.debug("Preparing startup message.")
            msg_startup = (
                MSG_STARTUP_SSH
                if (show_ssh_help or not webbrowser_available())
                else MSG_STARTUP
            )
            url = instance.url()
            host_ports = instance.host_ports()
            assert len(host_ports) > 0
            click.secho(
                msg_startup.format(
                    url=url,
                    home_mount=instance.profile.home_mount,
                    system_user=instance.profile.system_user,
                    user=getpass.getuser(),
                    port=host_ports[0],
                    hostname=socket.getfqdn(),
                ).lstrip(),
                fg="green",
            )

            for extra_mount in profile.extra_mounts:
                source, target, mode = profile.parse_extra_mount(extra_mount)
                click.secho(
                    MSG_EXTRA_VOLUME.format(
                        source=source,
                        target=target,
                        mode=f"({mode})" if mode else "",
                    ).lstrip(),
                    fg="green",
                )

            if not no_browser and webbrowser_available():
                if click.confirm(
                    "Do you want to open AiiDAlab in the browser now?", default=True
                ):
                    click.launch(url)

        else:
            click.secho(
                "Use 'aiidalab-launch status' to check the AiiDAlab instance "
                "status and URL to open it.",
                fg="green",
            )


@cli.command()
@click.option(
    "--restart",
    is_flag=True,
    help="Restart the container in case that it is already running.",
)
@click.option(
    "--wait",
    default=120,
    show_default=True,
    help="Time to wait after startup until all services are up. Set to zero to not wait at all and immediately return.",
)
@click.option(
    "--pull/--no-pull",
    default=True,
    help=(
        "Specify whether to pull the configured image prior to the first start "
        "of the container."
    ),
    show_default=True,
)
@click.option(
    "--no-browser",
    is_flag=True,
    help=(
        "Do not open AiiDAlab in the browser after startup. "
        "(This is disabled by default if the wait time is set to zero.)"
    ),
)
@click.option(
    "--show-ssh-port-forwarding-help",
    "show_ssh_help",
    is_flag=True,
    help="Show guidance on SSH port forwarding.",
)
@click.option(
    "-f", "--force", is_flag=True, help="Ignore any warnings and start anyways."
)
@pass_app_state
@with_profile
def start(*args, **kwargs):
    """Start an AiiDAlab instance on this host."""
    asyncio.run(_async_start(*args, **kwargs))


@cli.command()
@click.option(
    "-r",
    "--remove",
    is_flag=True,
    help="Do not only stop the container, but also remove it.",
)
@click.option(
    "-t",
    "--timeout",
    type=click.INT,
    default=20,
    help="Wait this long for the instance to shut down.",
)
@pass_app_state
@with_profile
def stop(app_state, profile, remove, timeout):
    """Stop an AiiDAlab instance on this host."""
    instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
    status = asyncio.run(instance.status())
    if status not in (
        instance.AiidaLabInstanceStatus.DOWN,
        instance.AiidaLabInstanceStatus.CREATED,
        instance.AiidaLabInstanceStatus.EXITED,
    ):
        with spinner("Stopping AiiDAlab...", final="stopped."):
            instance.stop(timeout=timeout)
    if remove:
        with spinner("Removing container..."):
            instance.remove()


@cli.command("logs")
@click.option("-f", "--follow", is_flag=True, help="Follow log output.")
@pass_app_state
@with_profile
def logs(app_state, profile, follow):
    """Show the logs of a running AiiDAlab instance."""
    instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
    try:
        for chunk in instance.logs(stream=True, follow=follow):
            click.echo(chunk, nl=False)
    except RuntimeError:
        raise click.ClickException("AiiDAlab instance was not created.")
    except KeyboardInterrupt:
        pass


@cli.command("status")
@pass_app_state
def status(app_state):
    """Show AiiDAlab instance status and entry point."""
    client = app_state.docker_client

    # Collect status for each profile
    headers = ["Profile", "Container", "Status", "Mount", "URL"]
    rows = []
    instances = [
        AiidaLabInstance(client=client, profile=profile)
        for profile in app_state.config.profiles
    ]

    async def fetch_status(instances):
        results = await asyncio.gather(*(instance.status() for instance in instances))
        return list(zip(instances, results))

    with spinner("Collecting status info...", delay=0.5):
        for instance, instance_status in asyncio.run(fetch_status(instances)):
            rows.append(
                [
                    instance.profile.name,
                    instance.profile.container_name(),
                    {
                        AiidaLabInstance.AiidaLabInstanceStatus.STARTING: "starting..."
                    }.get(instance_status, instance_status.name.lower()),
                    instance.profile.home_mount,
                    (
                        instance.url()
                        if instance_status is AiidaLabInstance.AiidaLabInstanceStatus.UP
                        else ""
                    ),
                ]
            )

    click.echo(tabulate(rows, headers=headers))


@cli.command()
@click.argument("cmd", nargs=-1)
@click.option("--privileged", is_flag=True)
@click.option("--forward-exit-code", is_flag=True)
@click.option(
    "--wait/--no-wait",
    help="Wait on AiiDAlab services to get ready before sending the command.",
    show_default=True,
    default=True,
)
@click.pass_context
@with_profile
def exec(ctx, profile, cmd, privileged, forward_exit_code, wait):
    """Directly execute a command on an AiiDAlab instance.

    For example, to get a list of all installed aiidalab applications, run:

        aiidalab-launch exec -- aiidalab list

    """
    app_state = ctx.ensure_object(ApplicationState)
    instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
    try:
        if wait:
            with spinner("Waiting for AiiDAlab to get ready...", delay=0.5):
                asyncio.run(asyncio.wait_for(instance.wait_for_services(), timeout=60))
        with spinner("Send command to container...", delay=1.0):
            exec_id = instance.exec_create(" ".join(cmd), privileged=privileged)
    except (RuntimeError, asyncio.TimeoutError):
        raise click.ClickException("AiiDAlab instance is not available. Is it running?")

    output = app_state.docker_client.api.exec_start(exec_id, stream=True)
    for chunk in output:
        click.echo(chunk.decode(), nl=False)

    result = app_state.docker_client.api.exec_inspect(exec_id)
    if result["ExitCode"] != 0:
        if forward_exit_code:
            ctx.exit(result["ExitCode"])
        else:
            ctx.fail(f"Command failed with exit code: {result['ExitCode']}")


@cli.command()
@click.option("--yes", is_flag=True, help="Do not ask for confirmation.")
@with_profile
@pass_app_state
def reset(app_state, profile, yes):
    """Reset an AiiDAlab instance.

    This function removes all user-installed apps and generated data. Use with
    caution.
    """
    # Check (and abort) in case that the instance is running.
    instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
    status = asyncio.run(instance.status())
    if status not in (
        instance.AiidaLabInstanceStatus.DOWN,
        instance.AiidaLabInstanceStatus.CREATED,
        instance.AiidaLabInstanceStatus.EXITED,
    ):
        raise click.ClickException(
            f"The instance associated with profile '{profile.name}' "
            "is still running. Please stop it prior to reset."
        )

    click.secho(
        f"Resetting instance for profile '{profile.name}'. This action cannot be undone!",
        err=True,
        fg="red",
    )

    if not yes:
        confirm_with_value(
            profile.name, "Please enter the name of the profile to continue", abort=True
        )

    click.echo("Removing container and associated (data) volumes.")
    instance.remove(conda=True, data=True)


if __name__ == "__main__":
    cli()



import asyncio
import logging
import re
import webbrowser
from contextlib import contextmanager
from pathlib import Path, PosixPath, PurePosixPath, WindowsPath
from textwrap import wrap
from threading import Event, Thread, Timer
from typing import Any, AsyncGenerator, Generator, Iterable, Optional, Union

import click
import click_spinner
import docker
import requests
from packaging.version import Version, parse
from requests_cache import CachedSession

from .core import LOGGER

MSG_UNABLE_TO_COMMUNICATE_WITH_CLIENT = (
    "Unable to communicate with docker on this host. This error usually indicates "
    "that Docker is either not installed on this system, that the docker service is "
    "not started, or that the installation is ill-configured.  Please follow the "
    "instructions at https://docs.docker.com/get-docker/ to install and start "
    "docker."
)


SESSION = CachedSession(
    "http_cache",
    backend="sqlite",
    use_cache_dir=True,
    expire_after=3600,  # 1hr
    stale_if_error=True,
)


@contextmanager
def spinner(
    msg: Optional[str] = None, final: Optional[str] = "done.", delay: float = 0
) -> Generator[None, None, None]:
    """Display spinner only after an optional initial delay."""

    def spin() -> None:

        # Don't show spinner if verbose output is enabled
        level = logging.getLogger().getEffectiveLevel()
        show_spinner = (
            True if level == logging.NOTSET or level >= logging.ERROR else False
        )
        if msg:
            newline = not show_spinner
            click.echo(f"{msg.rstrip()} ", nl=newline, err=True)

        if show_spinner:
            with click_spinner.spinner():  # type: ignore
                stop.wait()
            click.echo(final if (completed.is_set() and msg) else " ", err=True)
        else:
            stop.wait()

    stop = Event()
    completed = Event()
    timed_spinner = Timer(delay, spin)
    timed_spinner.start()
    try:
        yield
        # Try to cancel the timer if still possible.
        timed_spinner.cancel()
        # Set the completed event since there was no exception, indicating that
        # the waited on operation completed successfully.
        completed.set()
    finally:
        stop.set()
        timed_spinner.join()


def get_docker_client(*args, **kwargs) -> docker.client.DockerClient:
    try:
        with spinner("Connecting to docker host...", delay=0.2):
            return docker.from_env(*args, **kwargs)
    except docker.errors.DockerException as error:
        click.secho(
            "\n".join(wrap(MSG_UNABLE_TO_COMMUNICATE_WITH_CLIENT)),
            fg="yellow",
            err=True,
        )
        raise click.ClickException(f"Failed to communicate with Docker client: {error}")


def webbrowser_available() -> bool:
    """Check whether a webbrowser is available.

    Useful to provide more targeted user feedback, e.g., when AiiDAlab is
    executed on a headless node via ssh.
    """

    try:
        webbrowser.get()
    except webbrowser.Error:
        return False
    else:
        return True


# Adapted from: https://stackoverflow.com/a/62297994
def _async_wrap_iter(it: Iterable) -> AsyncGenerator[Any, None]:
    """Wrap blocking iterator into an asynchronous one"""
    loop = asyncio.get_event_loop()
    q: asyncio.Queue = asyncio.Queue(1)
    exception = None
    _END = object()

    async def yield_queue_items() -> AsyncGenerator[Any, None]:
        while True:
            next_item = await q.get()
            if next_item is _END:
                break
            yield next_item
        if exception is not None:
            # the iterator has raised, propagate the exception
            raise exception

    def iter_to_queue() -> None:
        nonlocal exception
        try:
            for item in it:
                # This runs outside the event loop thread, so we
                # must use thread-safe API to talk to the queue.
                asyncio.run_coroutine_threadsafe(q.put(item), loop).result()
        except Exception as e:
            exception = e
        finally:
            asyncio.run_coroutine_threadsafe(q.put(_END), loop).result()

    Thread(target=iter_to_queue).start()
    return yield_queue_items()


def get_latest_version(timeout: float = 0.1) -> Optional[Version]:
    """Determine the latest released version (on PyPI) of this tool."""
    try:
        req = SESSION.get(
            "https://pypi.python.org/pypi/aiidalab-launch/json", timeout=timeout
        )
        req.raise_for_status()
        releases = sorted(
            (
                version
                for version, release in req.json()["releases"].items()
                if not all(r["yanked"] for r in release)
            ),
            key=parse,
        )
        return parse(releases[-1]) if releases else None
    except (requests.exceptions.Timeout, requests.exceptions.ReadTimeout):
        LOGGER.debug("Timed out while requesting latest version.")
        return None
    except OSError as error:
        LOGGER.debug(f"Error while requesting latest version: {error}")
        return None


def confirm_with_value(value: str, text: str, abort: bool = False) -> bool:
    if click.prompt(text, default="", show_default=False) == value:
        return True
    elif abort:
        raise click.Abort
    else:
        return False


def get_docker_mount(
    container: docker.models.containers.Container, destination: PurePosixPath
) -> docker.types.Mount:
    try:
        mount = [
            mount
            for mount in container.attrs["Mounts"]
            if mount["Destination"] == str(destination)
        ][0]
    except IndexError:
        raise ValueError(f"No mount point for {destination}.")
    return mount


def is_volume_readonly(
    container: docker.models.containers.Container, destination: PurePosixPath
) -> bool:
    mount = get_docker_mount(container, destination)
    return not mount["RW"]


def docker_mount_for(
    container: docker.models.containers.Container, destination: PurePosixPath
) -> Union[Path, str]:
    """Identify the Docker mount bind path or volume for a given destination."""
    mount = get_docker_mount(container, destination)
    if mount["Type"] == "bind":
        docker_root = PurePosixPath("/host_mnt")
        docker_path = PurePosixPath(mount["Source"])
        try:  # Windows
            drive = docker_path.relative_to(docker_root).parts[0]
            return WindowsPath(
                f"{drive}:",
                docker_path.root,
                docker_path.relative_to(docker_root, drive),
            )
        except ValueError:  # Linux
            return PosixPath(docker_path)
        except NotImplementedError:  # OS-X
            return PosixPath(docker_root.root, docker_path.relative_to(docker_root))
    elif mount["Type"] == "volume":
        return mount["Name"]
    else:
        raise RuntimeError("Unexpected mount type.")


def get_docker_env(container: docker.models.containers.Container, env_name: str) -> str:
    re_pattern = f"{re.escape(env_name)}=(?P<value>.*)"
    try:
        for item in container.attrs["Config"]["Env"]:
            match = re.search(re_pattern, item)
            if match:
                return match.groupdict()["value"]
    except KeyError:
        pass
    raise KeyError(env_name)
